#include <assert.h>

#include <QDir>
#include <QDomDocument>
#include <QFontMetrics>
#include <QGraphicsGridLayout>
#include <QGraphicsScene>
#include <QGraphicsProxyWidget>
#include <QLabel>
#include <QPainter>
#include <QString>
#include <QTimer>
#include <QtXmlPatterns>

#include <KTextEdit>

#include <Plasma/TextEdit>
#include <Plasma/Theme>

#include "splunk_status.h"

K_EXPORT_PLASMA_APPLET(splunk_status, StatusApplet)

static const QString DEFAULT_SPLUNK_USER = "admin";
static const QString DEFAULT_SPLUNK_PASS = "changeme";

#define COLUMN_ID   0
#define COLUMN_INFO 1

#define PROC_START_TIMEOUT     500      // milliseconds.
#define STATUS_CHECK_INTERVAL  10000    // milliseconds.

StatusApplet::StatusApplet(QObject* parent, const QVariantList& args)
    : Plasma::Applet(parent, args), _interval(STATUS_CHECK_INTERVAL), _layout(NULL),
      _timer(NULL)
{
  resize(600, 200);
}

StatusApplet::~StatusApplet()
{
  delete _timer;
}

void StatusApplet::init()
{
  loadSettings();

  _layout = new QGraphicsGridLayout();
  _layout->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Preferred);
  setLayout(_layout);
  
  unsigned i;
  QMap<QString, InstanceInfo>::iterator it;
  for (i = 0, it = _instanceInfoMap.begin(); _instanceInfoMap.end() != it; ++it, ++i) {
    // add server id.
    Plasma::TextEdit* textName = new Plasma::TextEdit();
    textName->setMaximumHeight(40); // reasonably small.
    textName->setMaximumWidth(150); // because my stretchfactors appear to do nothing...
    textName->setText(QString("<font color=grey>") + it.key() + "</font>");
    // shrink this column.
    QSizePolicy shrinkPolicy;
    shrinkPolicy.setHorizontalPolicy(QSizePolicy::Maximum);
    textName->setSizePolicy(shrinkPolicy);
    // set colors and add.
    textName->setStyleSheet("KTextEdit { background-color: black }");
    _layout->addItem(textName, i, COLUMN_ID);

    // add space for results.
    Plasma::TextEdit* textMsgs = new Plasma::TextEdit();
    textMsgs->setMaximumHeight(40); // reasonbly small.
    textMsgs->setText(QString("<font color=yellow>status not yet checked</font>"));
    // expand message column as much as possible.
    QSizePolicy growPolicy;
    growPolicy.setHorizontalPolicy(QSizePolicy::MinimumExpanding);
    textMsgs->setSizePolicy(growPolicy);
    // set colors and add.
    textMsgs->setStyleSheet("KTextEdit { background-color: black }");
    _layout->addItem(textMsgs, i, COLUMN_INFO);
    // make results column bigger.  this appears to have no effect. :(
    _layout->setColumnStretchFactor(COLUMN_ID,   1);
    _layout->setColumnStretchFactor(COLUMN_INFO, 9);
  }

  updateMessages();

  _timer = new QTimer(this);
  connect(_timer, SIGNAL(timeout()), this, SLOT(updateMessages()));
  _timer->start(_interval);
}

static void parse_messages_from_xml(QByteArray xml, QStringList& messages)
{
  // put XML into a QIODevice.
  QBuffer xmlData;
  xmlData.setData(xml);
  xmlData.open(QBuffer::ReadWrite);

  // build xpath query with splunk's namespace info.
  QXmlQuery query;
  query.bindVariable("data", &xmlData);
  query.setQuery("declare namespace a='http://www.w3.org/2005/Atom';"
                 "declare namespace s='http://dev.splunk.com/ns/rest';"
                 // choose only the s:key nodes that match their entry node's title.
                 "doc($data)/a:feed/a:entry/a:content/s:dict/string(s:key[(../../../a:title = @name)])");

  if (!query.evaluateTo(&messages))
    messages << "Parsing of status failed.";

  for (QStringList::iterator it = messages.begin(); messages.end() != it;) {
    QStringList::iterator currItem = it++;
    if (0 == currItem->length())
      messages.erase(currItem);
  }
}


bool StatusApplet::getMessages(const QString& id)
{
  QMap<QString, InstanceInfo>::iterator it = _instanceInfoMap.find(id);
  assert(_instanceInfoMap.end() != it);

  const QString& uri = it.key();
  InstanceInfo& info = it.value();

  info.messages.clear();

  // figure out which user/pass to use.
  QString userPass = info.user + ":" + info.pass;

  // build args.
  QStringList args;
  args << "_internal"
       << "call"
       << "/admin/messages"
       << "-auth" << userPass; // this is OK even in the free version.
  if (!uri.isEmpty())
    args << "-uri" << (QString("http") + (info.useSSL ? "s" : "") + "://" + uri);

  // run w/ async notification of completion.
  RestProcess *proc = new RestProcess(uri);
  connect(proc, SIGNAL(finished(int, QProcess::ExitStatus)), this, SLOT(procFinished(int, QProcess::ExitStatus)));
  proc->start(_binPath, args);
  if (!proc->waitForStarted(PROC_START_TIMEOUT)) { // half sec timeout should be plenty.
    delete proc;
    return false;
  }

  return true;
}

void StatusApplet::procFinished(int exitCode, QProcess::ExitStatus exitStatus)
{
  RestProcess* caller = static_cast<RestProcess*>(sender());
  assert(0 != caller); // we have to be called via a signal.

  const QString& id = caller->getInstanceId();

  // since i'm being lazy, the row for this host is tied directly to the order in the map.  look it up.
  unsigned row;
  QMap<QString, InstanceInfo>::iterator it;
  for (row = 0, it = _instanceInfoMap.begin(); _instanceInfoMap.end() != it; ++row, ++it)
    if (id == it.key())
      break;

  assert(_instanceInfoMap.end() != it);

  InstanceInfo& info = it.value();

  // non-zero return indicates error.  non-normal exit indicates process crash.
  if (0 != exitCode || QProcess::NormalExit != exitStatus) {
    info.messages << caller->readAllStandardError();
  }
  else {
    QByteArray output = caller->readAllStandardOutput();
    parse_messages_from_xml(output, info.messages);
  }

  updateDisplay(row, info.messages);

  delete(caller);
}

void StatusApplet::updateDisplay(const unsigned& row, const QStringList& messages)
{
  Plasma::TextEdit* idItem  = static_cast<Plasma::TextEdit*>(_layout->itemAt(row, COLUMN_ID));
  Plasma::TextEdit* msgItem = static_cast<Plasma::TextEdit*>(_layout->itemAt(row, COLUMN_INFO));
  KTextEdit* msgText = static_cast<KTextEdit*>(static_cast<Plasma::TextEdit*>(msgItem)->widget());
  if (messages.isEmpty())
    msgText->setText("<font color=green>No messages :)</font>");
  else {
    msgText->setText(QString("<font color=red>") + messages.join("<font color=yellow> ::: </font>") + "</font>");
  }

  idItem->setMaximumHeight( msgText->document()->size().height() + 10); // why +10?
  msgItem->setMaximumHeight(msgText->document()->size().height() + 10); // hell if i know...
  resize(effectiveSizeHint(Qt::PreferredSize));

}

void StatusApplet::updateMessages()
{
  for (QMap<QString, InstanceInfo>::iterator it = _instanceInfoMap.begin(); _instanceInfoMap.end() != it; ++it)
    getMessages(it.key());
}

void StatusApplet::loadSettings()
{
  QString settingsPath = QDir::homePath() + "/.kde/share/config/splunkstatusrc"; // cross-platform safe.
  QFile settingsFile(settingsPath);
  if (!settingsFile.open(QIODevice::ReadOnly | QIODevice::Text)) {
    fprintf(stderr, "Unable to open settings file at: %s.", settingsPath.toAscii().constData());
    _instanceInfoMap["COULD NOT READ SETTINGS FILE"]; // HELP, ABUSE.
    return;
  }

  QString section;
  while (!settingsFile.atEnd()) {
    QByteArray line = settingsFile.readLine().trimmed();
    // new stanza name.
    if (line.startsWith('[')) {
      line = line.mid(1); // strip 1st [.
      if (line.endsWith(']'))
        line = line.mid(0, line.length() - 1); // strip last ].
      section = line;
    }
    // an actual setting.
    else {
      QList<QByteArray> postSplit = line.split('='); // fails on >1 = sign.  don't really care here.
      if (2 > postSplit.size()) { // needs to be key = value.
        continue; 
      }
      if ("settings" == section) {
        if ("cmdPath" == postSplit[0].trimmed()) {
          _binPath = postSplit[1].trimmed(); // path to splunk binary.
        }
      }
      // [servers]
      // 10.0.1.2:9089 =
      // 10.1.2.3:8089 = admin,changeme,ssl
      // 10.2.3.4:8089 = admin,changeme,nossl
      // 10.3.4.5:9089 = admin,changeme        <-- implies ssl.
      else if ("servers" == section) {
        InstanceInfo& info = _instanceInfoMap[postSplit[0].trimmed()]; // store uri.
        QList<QByteArray> serverSettings = postSplit[1].split(',');
        if (serverSettings.size() > 2)
          info.useSSL = ("ssl" == serverSettings[2].trimmed());
        if (serverSettings.size() > 1)
          info.pass   = serverSettings[1].trimmed();
        if (serverSettings.size() > 0)
          info.user   = serverSettings[0].trimmed();
      }
    }
  }
}


#include "splunk_status.moc"
